"use strict";
/******************************************************************************
 * Copyright 2022 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
Object.defineProperty(exports, "__esModule", { value: true });
exports.eager = exports.inject = void 0;
const merge_1 = require("./merge");
const isEager = Symbol();
const isRequested = Symbol();
// @ts-expect-error ts(2370)
function inject(...modules) {
    const module = modules.reduce(merge_1.merge, {});
    const container = proxify(module);
    initializeEagerServices(module, container);
    return container;
}
exports.inject = inject;
function eager(factory) {
    return (isEager in factory) ? factory : Object.assign(((ctr) => factory(ctr)), { [isEager]: true });
}
exports.eager = eager;
function initializeEagerServices(module, context) {
    (0, merge_1.keys)(module).forEach(key => {
        const t = module[key];
        (typeof t === 'function') ? ((isEager in t) && t(context)) : initializeEagerServices(t, context);
    });
}
function proxify(module, container, path = '') {
    const get = (obj, prop, proxy) => {
        const name = path + '[' + String(prop) + ']';
        if (obj[prop] === isRequested) {
            throw new Error('Cyclic dependency ' + name + '. See https://djinject.io/#cyclic-dependencies');
        }
        const ctr = container || proxy;
        const val = module[prop];
        return (prop in obj) ? obj[prop] : (prop in module) ? (obj[prop] = isRequested,
            obj[prop] = (typeof val === 'function') ? val(ctr) : proxify(val, ctr, name)) : undefined;
    };
    const proxy = new Proxy({}, {
        deleteProperty: () => false,
        get,
        getOwnPropertyDescriptor: (target, prop) => (get(target, prop, proxy), Object.getOwnPropertyDescriptor(target, prop)),
        has: (_, prop) => prop in module,
        ownKeys: () => Reflect.ownKeys(module)
    });
    return proxy;
}
//# sourceMappingURL=inject.js.map